import torch
import random

from typing import List, Dict, Any
from torch.utils.data import DataLoader

from ..pipeline import PipelineContext

from ...losses import get_loss
from ...utilities import get_logger
from ...optimizers import get_optimizer
from ...solvers import get_solver, Solver
from ...models import get_model, FineTuningModel
from ...datasets import get_dataset, get_embeddings_collate_function, BiotrainerDataset

logger = get_logger(__name__)


class TrainingFactory:
    @staticmethod
    def create_dataset(context: PipelineContext, split: List, mode: str, finetuning: bool = False) -> BiotrainerDataset:
        # Apply limited sample number
        limited_sample_size = context.config["limited_sample_size"]
        if mode == "train" and limited_sample_size and limited_sample_size > 0:
            logger.info(f"Using limited sample size of {limited_sample_size} for training dataset")
            split = random.sample(split,
                                  k=min(limited_sample_size, len(split)))

        random_masking = context.config.get("finetuning_config", {}).get("random_masking", False)
        mask_token = context.embedding_service._embedder.get_mask_token() if context.embedding_service is not None else None
        return get_dataset(samples=split,
                           finetuning=finetuning,
                           random_masking=random_masking,
                           mask_token=mask_token,
                           class_str2int=context.class_str2int
                           )

    @staticmethod
    def create_dataloader(context: PipelineContext, dataset, hyper_params: Dict,
                          finetuning: bool = False) -> DataLoader:
        # Create dataloader from dataset
        if finetuning:
            collate_fn = lambda batch: (
                [x[0] for x in batch], [x[1] for x in batch], [x[2] for x in batch], [len(x[1]) for x in batch])
        else:
            collate_fn = get_embeddings_collate_function(context.config["protocol"])
        return DataLoader(
            dataset=dataset, batch_size=hyper_params["batch_size"], shuffle=hyper_params["shuffle"], drop_last=False,
            collate_fn=collate_fn
        )

    @staticmethod
    def create_solver(context: PipelineContext, split_name: str, model, loss_function, optimizer,
                      hyper_params: Dict) -> Solver:
        return get_solver(protocol=context.config["protocol"], name=split_name, network=model, optimizer=optimizer,
                          loss_function=loss_function,
                          output_manager=context.output_manager,
                          device=context.config["device"],
                          number_of_epochs=hyper_params["num_epochs"],
                          patience=hyper_params["patience"], epsilon=hyper_params["epsilon"],
                          log_dir=hyper_params["log_dir"], n_classes=context.n_classes)

    @staticmethod
    def create_model_loss_optimizer(context: PipelineContext,
                                    hyper_params: Dict[str, Any]) -> (
            torch.nn.Module, torch.nn.Module, torch.nn.Module):
        # Initialize model
        model = get_model(n_classes=context.n_classes, n_features=context.n_features,
                          **hyper_params)

        if "finetuning_config" in context.config:
            model = FineTuningModel(embedding_service=context.embedding_service,
                                    downstream_model=model,
                                    collate_fn=get_embeddings_collate_function(context.config["protocol"]),
                                    protocol=context.config["protocol"],
                                    device=context.config["device"], )

        # Initialize loss function
        loss_function = get_loss(weight=context.class_weights, **hyper_params)

        # Initialize optimizer
        optimizer = get_optimizer(model_parameters=model.parameters(), **hyper_params)

        return model, loss_function, optimizer



from typing import Dict, Any, Tuple, List

from .config_option import ConfigOption, ConfigConstraints, ConfigKey

from ..protocols import Protocol


def get_default_cross_validation_config() -> Dict[str, Any]:
    return {"cross_validation_config": {"method": "hold_out"}}


def cross_validation_config(protocol: Protocol) -> Tuple[ConfigKey, List[ConfigOption]]:
    cv_category = "cross_validation"
    return ConfigKey.CROSS_VALIDATION, [
        ConfigOption(
            name="method",
            description="Define the method to use for cross-validation.",
            category=cv_category,
            required=True,
            default="hold_out",
            constraints=ConfigConstraints(
                type=str,
                allowed_values=["hold_out", "k_fold", "leave_p_out"]
            )
        ),
        ConfigOption(
            name="k",
            description="Define the number of folds to use for k-fold cross-validation.",
            category=cv_category,
            required=False,
            default=5,
            add_if=lambda config: config["method"] == "k_fold",
            constraints=ConfigConstraints(
                type=int,
                gte=2,
            )
        ),
        ConfigOption(
            name="stratified",
            description="Define whether to use stratified cross-validation, that tries to balance the classes "
                        "within the folds.",
            category=cv_category,
            required=False,
            default=False,
            add_if=lambda config: config["method"] == "k_fold",
            constraints=ConfigConstraints(
                type=bool,
            )
        ),
        ConfigOption(
            name="repeat",
            description="Define the number of times to repeat each fold. Repeating the cross-validation process "
                        "multiple times can lead to more robust performance estimates.",
            category=cv_category,
            required=False,
            default=1,
            add_if=lambda config: config["method"] == "k_fold",
            constraints=ConfigConstraints(
                type=int,
                gte=1
            )
        ),
        ConfigOption(
            name="nested",
            description="Define whether to use nested cross-validation for hyperparameter optimization.",
            category=cv_category,
            required=False,
            default=False,
            add_if=lambda config: config["method"] == "k_fold",
            constraints=ConfigConstraints(
                type=bool,
            )
        ),
        ConfigOption(
            name="nested_k",
            description="Define the number of folds to use for nested hyperparameter optimization.",
            category=cv_category,
            required=False,
            constraints=ConfigConstraints(
                type=int,
                gte=2
            )
        ),
        ConfigOption(
            name="search_method",
            description="Define the search method to use for nested cross-validation hyperparameter optimization.",
            category=cv_category,
            required=False,
            constraints=ConfigConstraints(
                type=str,
                allowed_values=["random_search", "grid_search"],
            )
        ),
        ConfigOption(
            name="n_max_evaluations_random",
            description="Limit the number of hyperparameter combinations to evaluate during "
                        "random search hyperparameter optimization.",
            category=cv_category,
            required=False,
            constraints=ConfigConstraints(
                type=int,
                gte=2
            )
        ),
        ConfigOption(
            name="choose_by",
            description="Specify which evaluation metric to use when selecting the best model "
                        "during cross validation.",
            category=cv_category,
            required=False,
            default="loss",
            constraints=ConfigConstraints(
                type=str,
                allowed_values=["loss", "accuracy", "precision", "recall"]  # TODO Add all metrics
            )
        ),
        ConfigOption(
            name="p",
            description="Define the number of samples to leave out in each iteration of leave-p-out cross-validation.",
            category=cv_category,
            required=False,
            default=5,
            add_if=lambda config: config["method"] == "leave_p_out",
            constraints=ConfigConstraints(
                type=int,
                gte=1
            )
        )
    ]



from typing import Dict, Any


def deprecated_config_keys(config_dict: Dict[str, Any]) -> Dict[str, str]:
    deprecated_keys_with_replacement = {"sequence_file": "input_file",
                                        "labels_file": "input_file",
                                        "mask_file": "input_file",
                                        }

    contained_deprecated_keys = {}
    for key in config_dict.keys():
        if key in deprecated_keys_with_replacement:
            contained_deprecated_keys[key] = deprecated_keys_with_replacement[key]
    return contained_deprecated_keys


from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, Type, Any, List, Tuple, Callable, Dict

from ..protocols import Protocol


@dataclass
class ConfigOption:
    name: str
    required: bool
    description: str
    category: str
    is_file_option: bool = False
    allow_hyperparameter_optimization: bool = False
    add_if: Optional[Callable[[Dict[str, Any]], bool]] = None
    default: Optional[Any] = None
    constraints: Optional[ConfigConstraints] = None

    def to_dict(self):
        return {
            "name": self.name,
            "required": self.required,
            "description": self.description,
            "category": self.category,
            "is_file_option": self.is_file_option,
            "allow_hyperparameter_optimization": self.allow_hyperparameter_optimization,
            "default": self.default,
            "constraints": self.constraints.to_dict() if self.constraints else {},
        }

@dataclass
class ConfigConstraints:
    type: Optional[Type] = None
    allowed_values: Optional[List[Any]] = None
    allowed_formats: Optional[List[str]] = None
    allowed_protocols: Optional[List[Protocol]] = field(default_factory=lambda: Protocol.all())
    custom_validator: Optional[Callable[[Any], Tuple[bool, str]]] = None  # Function that returns validation and error
    lt: Optional[int] = None
    lte: Optional[int] = None
    gt: Optional[int] = None
    gte: Optional[int] = None

    def to_dict(self):
        return {
            "type": str(self.type.__name__),
            "allowed_values": self.allowed_values,
            "allowed_formats": self.allowed_formats,
            "allowed_protocols": [protocol.name for protocol in self.allowed_protocols],
            "lt": self.lt,
            "lte": self.lte,
            "gt": self.gt,
            "gte": self.gte,
        }

class ConfigKey(Enum):
    ROOT = ""
    HF_DATASET = "hf_dataset"
    CROSS_VALIDATION = "cross_validation_config"
    FINETUNING = "finetuning_config"

    @staticmethod
    def all_subconfig_keys() -> List[ConfigKey]:
        return [key for key in ConfigKey if key != ConfigKey.ROOT]



from copy import deepcopy

import torch

from typing import Union

from ..pipeline import PipelineContext, PipelineStep
from ..pipeline.pipeline_step import PipelineStepType

from ..target_manager import TargetManager

from ...protocols import Protocol
from ...utilities import get_logger, EmbeddingDatasetSample

logger = get_logger(__name__)


class DataLoadingStep(PipelineStep):

    def get_step_type(self) -> PipelineStepType:
        return PipelineStepType.DATA_LOADING

    @staticmethod
    def _get_class_weights(context: PipelineContext, target_manager: TargetManager) -> Union[
        None, torch.FloatTensor]:
        protocol = context.config["protocol"]
        # Get x_to_class specific logs and weights
        class_weights = None
        if protocol in Protocol.classification_protocols():
            context.output_manager.add_derived_values(
                {'class_int2str': target_manager.class_int2str,
                 'class_str2int': target_manager.class_str2int}
            )
            context.class_str2int = target_manager.class_str2int

            # Compute class weights to pass as bias to model if option is set
            class_weights = target_manager.compute_class_weights()
            if class_weights is not None:
                computed_class_weights = {class_index: class_value.item() for
                                          class_index, class_value in enumerate(class_weights)}
                context.output_manager.add_derived_values({'computed_class_weights': computed_class_weights})

        return class_weights

    def process(self, context: PipelineContext) -> PipelineContext:
        id2emb = context.id2emb
        assert id2emb is not None and len(id2emb) > 0, f"id2emb cannot be None or empty at the data loading step!"
        first_embedding = next(iter(context.id2emb.values()))

        # TARGETS => DATASETS
        target_manager = TargetManager(protocol=context.config["protocol"], input_data=context.input_data,
                                       ignore_file_inconsistencies=context.config["ignore_file_inconsistencies"],
                                       cross_validation_method=context.config["cross_validation_config"]["method"],
                                       interaction=context.config.get("interaction"))


        finetuning = "finetuning_config" in context.config
        if finetuning:
            train_dataset, val_dataset, test_datasets, prediction_dataset = target_manager.get_sequence_datasets()
            baseline_test_datasets = {test_name: [
                EmbeddingDatasetSample(seq_id=data_point.seq_id, embedding=context.id2emb[data_point.seq_id],
                                       target=data_point.target) for data_point in test_data] for test_name, test_data
                                      in test_datasets.items()}
            random_masking = context.config.get("finetuning_config", {}).get("random_masking", False)
            if random_masking:
                logger.info("Random masking is enabled. All targets will be overwritten with masked sequence targets.")
        else:
            train_dataset, val_dataset, test_datasets, prediction_dataset = target_manager.get_embedding_datasets(
                context.id2emb)
            baseline_test_datasets = deepcopy(test_datasets)

        del context.id2emb  # No longer required and should not be used later in the routine
        context.id2emb = None

        # LOG COMMON VALUES FOR ALL k-fold SPLITS:
        context.n_features = first_embedding.shape[-1]  # Last position in shape is always embedding dim
        context.n_classes = target_manager.number_of_outputs

        logger.info(f"Number of features: {context.n_features}")
        logger.info(f"Number of outputs (i.e. classes, equals 1 for regression): {context.n_classes}")

        context.output_manager.add_derived_values({
            'n_features': context.n_features,
            'n_testing_ids': sum(len(test_dataset) for test_dataset in test_datasets.values()),
            'n_classes': context.n_classes,
        })

        # Store datasets
        context.train_dataset = train_dataset
        context.val_dataset = val_dataset
        context.test_datasets = test_datasets
        context.baseline_test_datasets = baseline_test_datasets
        context.prediction_dataset = prediction_dataset

        # CLASS WEIGHTS
        context.class_weights = self._get_class_weights(context=context, target_manager=target_manager)

        context.target_manager = target_manager
        return context



from __future__ import annotations

import os
import shutil
import requests

from tqdm import tqdm
from pathlib import Path
from dataclasses import dataclass
from appdirs import user_cache_dir
from abc import ABC, abstractmethod
from typing import List, Optional, Union


class AutoEvalDataHandler(ABC):

    def download_data(self, data_dir: Path) -> None:
        """
        Download and extract a data archive from a list of URLs, using them as fallbacks
        Urls can be a single URL or list of URLs to download the data from (will be tried in order)

        Args:
            data_dir: Directory to extract the data to
        """
        urls = self.get_download_urls()

        zip_file = data_dir.with_suffix('.zip')
        headers = {
            'Accept': 'application/zip, application/octet-stream',
            'User-Agent': 'biotrainer/autoeval'
        }

        for i, url in enumerate(urls, 1):
            try:
                print(f"Attempting download from {url} (attempt {i}/{len(urls)})..")
                response = requests.get(url, headers=headers, stream=True)
                response.raise_for_status()

                total_size = int(response.headers.get('content-length', 0))
                block_size = 8192  # 8 KB

                with open(zip_file, "wb") as f, tqdm(
                        desc="Downloading data",
                        total=total_size,
                        unit='iB',
                        unit_scale=True,
                        unit_divisor=1024,
                ) as progress_bar:
                    for data in response.iter_content(block_size):
                        size = f.write(data)
                        progress_bar.update(size)

                print("Unpacking data archive..")
                shutil.unpack_archive(zip_file, data_dir)

                # Remove the zip file after successful extraction
                zip_file.unlink()

                print("Data downloaded and unpacked successfully!")
                return  # Success - exit the function

            except Exception as e:
                print(f"Failed to download from {url}: {e}")
                if zip_file.exists():
                    zip_file.unlink()

                # If this was the last URL, raise the exception
                if i == len(urls):
                    print("All download attempts failed")
                    raise Exception("Failed to download data from all provided URLs") from e

                # Otherwise, continue to the next URL
                print("Trying next fallback URL...")
                continue

    @staticmethod
    @abstractmethod
    def get_framework_name():
        raise NotImplementedError

    def get_framework_base_path(self, custom_storage_path: Optional[Union[str, Path]] = None) -> Path:
        if custom_storage_path:
            return Path(custom_storage_path) / self.get_framework_name()
        return Path(user_cache_dir('biotrainer')) / "autoeval" / self.get_framework_name()

    @staticmethod
    def get_preprocessed_file_dir_name(min_seq_length: int, max_seq_length: int):
        assert isinstance(min_seq_length, int) and isinstance(max_seq_length, int), \
            f"{min_seq_length} and {max_seq_length} must be integers"
        return f"preprocessed_{min_seq_length}_{max_seq_length}"

    @staticmethod
    @abstractmethod
    def get_download_urls():
        raise NotImplementedError

    @staticmethod
    def is_download_necessary(base_path: Path) -> bool:
        if not base_path.is_dir():
            raise ValueError(f"Given path {base_path} is not a directory!")
        return len(os.listdir(base_path)) == 0  # Directory is empty => Download

    @abstractmethod
    def preprocess(self, base_path: Path, min_seq_length: int, max_seq_length: int):
        raise NotImplementedError

    @abstractmethod
    def get_tasks(self, base_path: Path, min_seq_length: int, max_seq_length: int) -> List[AutoEvalTask]:
        """
        Get tasks to execute in the autoeval pipeline via biotrainer.

        This is the core method of the framework data handler.
        """
        raise NotImplementedError


@dataclass
class AutoEvalTask:
    name: str  # usually {FRAMEWORK_NAME}-{DATASET}-{optional: SPLIT_NAME}
    input_file: Path
    type: str
